"""
Módulo relacionado ao servidor que executará os comandos PowerShell
"""
import os
import re
import socket
import tempfile
import paramiko


def escapar_echo_cmd(conteudo):
    conteudo_escapado = re.sub(
        r'(&|\\|<|>|\^|\|)', r'^\1', conteudo, flags=re.RegexFlag.MULTILINE)
    conteudo_escapado = re.sub(r'\n', ';', conteudo_escapado)
    return conteudo_escapado


class ServidorAcesso:
    __PASTA_TEMPORARIA = 'vmm_temp'
    __ENCODE_CMD = 'cp850'
    __ENCODE_WINDOWS = 'iso-8859-1'
    __TIMEOUT_CONEXAO = 60
    __DEFAULT_SSH_PORT = 22

    @staticmethod
    def __get_caminho_arquivo(nome):
        return f'{ServidorAcesso.__PASTA_TEMPORARIA}/{os.path.basename(nome)}'

    def __init__(self, servidor, usuario, senha, servidor_vmm):
        self.servidor = servidor
        self.servidor_vmm = servidor_vmm
        self.usuario = usuario
        self.senha = senha

        self.conexao = None
        self.conexao_sftp = None
        self.__msg_erro_conexao = None
        self.__msg_erro_conexao_sftp = None

    def __del__(self):
        if self.conexao:
            self.conexao.close()
        if self.conexao_sftp:
            self.conexao_sftp.close()

    def get_msg_erro_conexao(self):
        if self.__msg_erro_conexao:
            return f'Não foi possível se conectar ao servidor de acesso. {self.__msg_erro_conexao}'
        return None

    def get_msg_erro_conexao_sftp(self):
        if self.__msg_erro_conexao_sftp:
            return ('Não foi possível abrir uma conexão sftp ' +
                    f'com o servidor de acesso. {self.__msg_erro_conexao_sftp}')
        return None

    def __is_conexao_ok(self):
        if self.conexao:
            self.conexao.close()
        self.__conectar()
        return not self.__msg_erro_conexao

    def __conectar(self):
        try:
            self.conexao = paramiko.SSHClient()
            self.conexao.load_system_host_keys()
            self.conexao.set_missing_host_key_policy(paramiko.AutoAddPolicy())
            self.conexao.connect(hostname=self.servidor,
                                 port=self.__DEFAULT_SSH_PORT,
                                 username=self.usuario,
                                 password=self.senha,
                                 banner_timeout=ServidorAcesso.__TIMEOUT_CONEXAO)
        except paramiko.AuthenticationException:
            self.__msg_erro_conexao = 'Usuário ou senha inválidos.'
        except paramiko.SSHException as ex:
            self.__msg_erro_conexao = f'Erro ao estabelecer conexão SSH: {ex}'
        except (socket.error, socket.timeout) as ex:
            self.__msg_erro_conexao = f'Erro de socket: {ex}'

    def __is_conexao_sftp_ok(self):
        if self.conexao_sftp:
            self.conexao_sftp.close()
        self.__conectar_sftp()
        return not self.__msg_erro_conexao_sftp

    def __conectar_sftp(self):
        try:
            paramiko_transport = paramiko.Transport(
                (self.servidor, self.__DEFAULT_SSH_PORT))
            paramiko_transport.connect(
                username=self.usuario, password=self.senha)
            self.conexao_sftp = paramiko.SFTPClient.from_transport(
                paramiko_transport)

        except paramiko.AuthenticationException:
            self.__msg_erro_conexao_sftp = 'Usuário ou senha inválidos.'
        except paramiko.SSHException as ex:
            self.__msg_erro_conexao_sftp = f'Erro ao estabelecer conexão SSH: {ex}'
        except (socket.error, socket.timeout) as ex:
            self.__msg_erro_conexao_sftp = f'Erro de socket: {ex}'

    def __executar_comando(self, cmd):
        if self.__is_conexao_ok():
            try:
                _, stdout, stderr = self.conexao.exec_command(cmd)
                stderr_msg = stderr.read().decode(ServidorAcesso.__ENCODE_CMD)
                if stdout.channel.recv_exit_status() != 0 or stderr_msg:
                    return False, stderr_msg

                return True, stdout.read().decode(ServidorAcesso.__ENCODE_CMD)
            except paramiko.SSHException as ex:
                return False, f"Erro ao executar comando '{cmd}': {ex}"
            except socket.error as ex:
                return False, f"Erro de socket ao executar '{cmd}': {ex}"
        else:
            return False, self.get_msg_erro_conexao()

    def __enviar_arquivo(self, nome_arquivo):
        resultado = self.__executar_comando(
            f'if not exist "{ServidorAcesso.__PASTA_TEMPORARIA}" '
            f'mkdir "{ServidorAcesso.__PASTA_TEMPORARIA}"')

        if not resultado[0]:
            return resultado

        if self.__is_conexao_sftp_ok():
            file_attrs = self.conexao_sftp.put(nome_arquivo,
                                               ServidorAcesso.__get_caminho_arquivo(
                                                   nome_arquivo),
                                               confirm=True)
            return True, file_attrs

        return False, self.get_msg_erro_conexao_sftp()

    def __excluir_arquivo(self, nome):
        self.__executar_comando('del {}'.format(
            ServidorAcesso.__get_caminho_arquivo(nome).replace('/', '\\')))

    def get_caminho_lockfile(self, agrupamento, nuvem):
        return self.__get_caminho_arquivo(f'{agrupamento}-{nuvem}.lock')

    def executar_script(self, nome, conteudo):
        try:
            arquivo_script = None

            if self.__is_conexao_ok():
                arquivo_script = tempfile.NamedTemporaryFile(
                    prefix=nome, suffix='.ps1', delete=True)
                arquivo_script.file.write(conteudo.encode(
                    ServidorAcesso.__ENCODE_WINDOWS))
                arquivo_script.flush()

                res_envio_arquivo = self.__enviar_arquivo(arquivo_script.name)
                if not res_envio_arquivo[0]:
                    return res_envio_arquivo

                caminho_arquivo = ServidorAcesso.__get_caminho_arquivo(
                    arquivo_script.name)
                resultado = self.__executar_comando(
                    f'powershell.exe -file {caminho_arquivo}')

                self.__excluir_arquivo(arquivo_script.name)

                return resultado

            return False, self.get_msg_erro_conexao()
        # pylint: disable=broad-except
        except Exception as ex:
            return False, f'Erro "{type(ex).__name__}" ao executar script: {ex}'
        finally:
            if arquivo_script:
                arquivo_script.close()



"""
Representação de um inventário
"""
import json
from vmm_manager.entidade.plano_execucao import PlanoExecucao
from vmm_manager.util.config import CAMPO_AGRUPAMENTO
from vmm_manager.infra.comando import Comando
from vmm_manager.scvmm.scregion import SCRegion


def json_handle_inventario(obj):
    if isinstance(obj, Inventario):
        return obj.to_dict()
    raise ValueError('Objeto precisa ser uma instância de inventário.')


class Inventario:

    @staticmethod
    def get_json(inventario_local, inventario_remoto, dados_completos=True):
        for nome_vm in inventario_remoto.vms:
            if nome_vm not in inventario_local.vms:
                # máquina órfã: não exibir
                continue

            inventario_remoto.vms[nome_vm].dados_ansible = \
                inventario_local.vms[nome_vm].dados_ansible

            # definindo tipo de impressão
            inventario_remoto.vms[nome_vm].to_json_dados_completos = dados_completos

        return True, json.dumps(inventario_remoto,
                                default=json_handle_inventario,
                                sort_keys=True, indent=4)

    def __init__(self, agrupamento, nuvem):
        self.agrupamento = agrupamento
        self.nuvem = nuvem
        self.vms = {}

        self.__regioes_por_letra_id = None

    def get_nome_no_regiao(self, regiao):
        if regiao in self.__regioes_por_letra_id or {}:
            return self.__regioes_por_letra_id[regiao].nome_no

        raise ValueError(f"Região '{regiao}' não possui nó definido.")

    def get_id_no_regiao(self, regiao):
        if regiao in self.__regioes_por_letra_id or {}:
            return self.__regioes_por_letra_id[regiao].id_no

        raise ValueError(f"Região '{regiao}' não possui nó definido.")

    def get_mapeamento_regioes(self):
        if self.__regioes_por_letra_id is None:
            ValueError('Mapeamento de regiões não definido.')

        return self.__regioes_por_letra_id

    def set_mapeamento_regioes(self, regioes_disponiveis):
        self.__regioes_por_letra_id = {}
        for regiao in regioes_disponiveis:
            self.__regioes_por_letra_id[regiao.letra_id] = regiao

    def calcular_plano_execucao(self, inventario_remoto):
        if (self.agrupamento != inventario_remoto.agrupamento
                or self.nuvem != inventario_remoto.nuvem):
            return False, 'Não é possível calcular o plano de execução \
                para inventários de agrupamento ou nuvem distintos.'

        plano_execucao = PlanoExecucao(self.agrupamento, self.nuvem)

        self.__add_acoes_criar_vms(inventario_remoto, plano_execucao)
        self.__add_acoes_execucao_excluir_vms(
            inventario_remoto, plano_execucao)

        self.__add_acoes_diferenca_vm(
            inventario_remoto, plano_execucao)
        self.__add_acoes_diferenca_discos_adicionais(
            inventario_remoto, plano_execucao)
        self.__add_acoes_diferenca_regiao(
            inventario_remoto, plano_execucao)

        return True, plano_execucao

    def gerar_plano_exclusao(self):
        plano_execucao = PlanoExecucao(self.agrupamento, self.nuvem)

        for maquina_virtual in self.vms.values():
            plano_execucao.acoes.append(maquina_virtual.get_acao_excluir_vm())

        return plano_execucao

    def is_vazio(self):
        return not self.vms

    def lista_nome_vms_str(self):
        return ','.join([f'"{nome_vm}"' for nome_vm in self.vms])

    def __validar_regras_locais(self):
        regioes = set()

        for maquina_virtual in self.vms.values():
            if maquina_virtual.regiao != SCRegion.REGIAO_PADRAO:
                regioes.add(maquina_virtual.regiao)

            if maquina_virtual.imagem is None:
                raise ValueError(
                    f'Imagem da VM {maquina_virtual.nome} não definida.')

            if maquina_virtual.qtde_cpu is None:
                raise ValueError(
                    f'Quantidade de CPUs da VM {maquina_virtual.nome} não definida.')

            if maquina_virtual.qtde_ram_mb is None:
                raise ValueError(
                    f'Quantidade de memória da VM {maquina_virtual.nome} não definida.')

            if maquina_virtual.get_qtde_rede_principal() != 1:
                raise ValueError(
                    f'VM {maquina_virtual.nome} deve ter exatamente uma rede principal.')

        regioes_validas = [chr(ord('A') + num)
                           for num in range(0, len(regioes))]
        for regiao in regioes:
            if regiao not in regioes_validas:
                raise ValueError(
                    f"Região '{regiao}' não é válida.")

    def validar(self, servidor_acesso):
        self.__validar_regras_locais()

        imagens = set()
        redes = set()
        regioes = set()

        for maquina_virtual in self.vms.values():
            imagens.add(maquina_virtual.imagem)

            if maquina_virtual.regiao != SCRegion.REGIAO_PADRAO:
                regioes.add(maquina_virtual.regiao)

            redes.update([rede.nome for rede in maquina_virtual.redes])

        cmd = Comando('validar_inventario', imagens=imagens,
                      nuvem=self.nuvem,
                      redes=redes,
                      servidor_vmm=servidor_acesso.servidor_vmm,
                      qtde_minima_regioes=len(regioes),
                      agrupamento=self.agrupamento,
                      lista_nome_vms_str=self.lista_nome_vms_str(),
                      campo_agrupamento=CAMPO_AGRUPAMENTO[0])
        _, msg = cmd.executar(servidor_acesso)
        if msg:
            raise ValueError(msg)

    def set_discos_vms(self, discos_adicionais):
        for nome_vm in discos_adicionais:
            self.vms[nome_vm].add_discos_adicionais(discos_adicionais[nome_vm])

    def __add_acoes_criar_vms(self, inventario_remoto, plano_execucao):
        vms_inserir = [
            nome_vm_local for nome_vm_local in self.vms
            if nome_vm_local not in inventario_remoto.vms
        ]
        for nome_vm in vms_inserir:
            plano_execucao.acoes.append(self.vms[nome_vm].get_acao_criar_vm())

    def __add_acoes_execucao_excluir_vms(self, inventario_remoto, plano_execucao):
        vms_excluir = [
            nome_vm_remoto for nome_vm_remoto in inventario_remoto.vms
            if nome_vm_remoto not in self.vms
        ]
        for nome_vm in vms_excluir:
            plano_execucao.acoes.append(
                inventario_remoto.vms[nome_vm].get_acao_excluir_vm())

    def __add_acoes_diferenca_discos_adicionais(self, inventario_remoto, plano_execucao):
        for nome_vm in self.vms:
            self.vms[nome_vm].add_acoes_diferenca_discos_adicionais(
                inventario_remoto.vms.get(nome_vm, None),
                plano_execucao)

    def __add_acoes_diferenca_regiao(self, inventario_remoto, plano_execucao):
        for nome_vm in self.vms:
            self.vms[nome_vm].add_acoes_diferenca_regiao(
                inventario_remoto.vms.get(nome_vm, None),
                plano_execucao, inventario_remoto)

    def __add_acoes_diferenca_vm(self, inventario_remoto, plano_execucao):
        for nome_vm in self.vms:
            if nome_vm in inventario_remoto.vms:
                self.vms[nome_vm].add_acoes_diferenca_vm(
                    inventario_remoto.vms[nome_vm], plano_execucao)

    def __eq__(self, other):
        return isinstance(other, Inventario) and (self.agrupamento == other.agrupamento
                                                  and self.nuvem == other.nuvem
                                                  and self.vms == other.vms)

    def __str__(self):
        return f'''
            agrupamento: {self.agrupamento}
            nuvem: {self.nuvem}
            vms: {self.vms}
            '''

    def to_dict(self):
        return {
            'agrupamento': self.agrupamento,
            'nuvem': self.nuvem,
            'vms': [self.vms[vm_name].to_dict() for vm_name in self.vms]
        }



"""
Representação de uma máquina virtual
"""
from vmm_manager.scvmm.enums import VMStatusEnum
from vmm_manager.entidade.vm_ansible import VMAnsible
from vmm_manager.entidade.acao import Acao
from vmm_manager.scvmm.scregion import SCRegion


class VM:
    def __init__(self, nome, descricao,
                 imagem, regiao,
                 qtde_cpu, qtde_ram_mb, redes,
                 id_vmm=None,
                 status=VMStatusEnum.EM_EXECUCAO,
                 no_regiao=None):
        self.nome = nome
        self.descricao = descricao if not descricao is None else ''
        self.imagem = imagem
        self.regiao = regiao
        self.qtde_cpu = qtde_cpu
        self.qtde_ram_mb = qtde_ram_mb
        self.redes = redes
        self.id_vmm = id_vmm
        self.status = status
        self.no_regiao = no_regiao

        self.dados_ansible = {}
        self.discos_adicionais = {}
        self.to_json_dados_completos = True

    def extrair_dados_ansible_dict(self, dict_ansible):
        for item in dict_ansible or {}:
            grupo = item.get('grupo')

            if grupo in self.dados_ansible:
                raise ValueError(
                    f"Grupo ansible '{grupo}' "
                    f"referenciado mais de uma vez para a VM '{self.nome}'.")

            ansible_grupo = VMAnsible(grupo)
            ansible_grupo.extrair_dados_vars_dict(
                item.get('vars'), self.nome)

            self.dados_ansible[grupo] = ansible_grupo

    def add_discos_adicionais(self, discos_adicionais):
        for disco_adicional in discos_adicionais:
            self.discos_adicionais[disco_adicional.arquivo] = disco_adicional

    def add_acoes_diferenca_discos_adicionais(self, vm_remota, plano_execucao):
        # discos a excluir
        if vm_remota:
            discos_excluir = [
                nome_disco_remoto for nome_disco_remoto in vm_remota.discos_adicionais
                if nome_disco_remoto not in self.discos_adicionais
            ]
            for nome_disco in discos_excluir:
                plano_execucao.acoes.append(
                    vm_remota.discos_adicionais[nome_disco].get_acao_excluir_disco(
                        vm_remota.id_vmm))

        # verificando discos atuais
        for nome_disco in self.discos_adicionais:
            # discos a criar
            if not vm_remota or not nome_disco in vm_remota.discos_adicionais:
                plano_execucao.acoes.append(
                    self.discos_adicionais[nome_disco].get_acao_criar_disco(self.nome))
            else:
                # discos a alterar
                plano_execucao.acoes.extend(
                    self.discos_adicionais[nome_disco].get_acoes_diferenca_disco(
                        vm_remota.discos_adicionais[nome_disco], vm_remota.id_vmm, self.nome))

    def add_acoes_diferenca_regiao(self, vm_remota,
                                   plano_execucao, inv_remoto):
        if (self.regiao != SCRegion.REGIAO_PADRAO
            and (not vm_remota
                 or (self.regiao != vm_remota.regiao or
                     inv_remoto.get_nome_no_regiao(self.regiao) != vm_remota.no_regiao))):
            plano_execucao.acoes.append(self.get_acao_mover_vm_regiao(
                inv_remoto.get_id_no_regiao(self.regiao)))

    def add_acoes_diferenca_vm(self, vm_remota, plano_execucao):
        # alteração da imagem é irreversível
        if self.imagem != vm_remota.imagem:
            plano_execucao.acoes.append(vm_remota.get_acao_excluir_vm())
            plano_execucao.acoes.append(self.get_acao_criar_vm())
            return

        # Alteração de rede é possível recuperar TODO #18
        if self.redes != vm_remota.redes:
            plano_execucao.acoes.append(vm_remota.get_acao_excluir_vm())
            plano_execucao.acoes.append(self.get_acao_criar_vm())
            return

        # alteração de descrição, cpu ou ram
        if (self.descricao != vm_remota.descricao
                or self.qtde_cpu != vm_remota.qtde_cpu
                or self.qtde_ram_mb != vm_remota.qtde_ram_mb):
            plano_execucao.acoes.append(
                self.get_acao_atualizar_vm(vm_remota.id_vmm))

    def get_qtde_rede_principal(self):
        return sum([1 for rede in self.redes if rede.principal])

    def get_rede_principal(self):
        return next((rede.nome for rede in self.redes if rede.principal), None)

    def get_acao_criar_vm(self):
        return Acao(Acao.ACAO_CRIAR_VM,
                    nome=self.nome,
                    descricao=self.descricao,
                    imagem=self.imagem,
                    regiao=self.regiao,
                    qtde_cpu=self.qtde_cpu,
                    qtde_ram_mb=self.qtde_ram_mb,
                    redes=[rede.nome for rede in self.redes],
                    rede_principal=self.get_rede_principal()
                    )

    def get_acao_excluir_vm(self):
        return Acao(Acao.ACAO_EXCLUIR_VM,
                    id_vmm=self.id_vmm)

    def get_acao_mover_vm_regiao(self, id_no_regiao):
        return Acao('mover_vm_regiao',
                    nome_vm=self.nome,
                    id_no_regiao=id_no_regiao,
                    regiao=self.regiao)

    def get_acao_atualizar_vm(self, id_vm_remota):
        return Acao('atualizar_vm',
                    id_vm=id_vm_remota,
                    descricao=self.descricao,
                    qtde_cpu=self.qtde_cpu,
                    qtde_ram_mb=self.qtde_ram_mb)

    def __hash__(self):
        return hash(self.nome)

    def __eq__(self, other):
        return isinstance(other, VM) and (self.nome == other.nome
                                          and self.imagem == other.imagem
                                          and self.regiao == other.regiao
                                          and self.qtde_cpu == other.qtde_cpu
                                          and self.qtde_ram_mb == other.qtde_ram_mb
                                          and self.redes == other.redes
                                          and self.status == other.status)

    def __repr__(self):
        return f'''
                nome: {self.nome}
                descricao: {self.descricao}
                imagem: {self.imagem}
                regiao: {self.regiao}
                qtde_cpu: {self.qtde_cpu}
                qtde_ram_mb: {self.qtde_ram_mb}
                redes: {self.redes}
                id_vmm: {self.id_vmm}
                status: {self.status}
                no_regiao: {self.no_regiao}
                ansible: {self.dados_ansible}
                discos_adicionais: {self.discos_adicionais}
                '''

    def to_dict(self):
        dict_objeto = {
            'nome': self.nome,
            'descricao': self.descricao,
            'imagem': self.imagem,
            'regiao': self.regiao,
            'qtde_cpu': self.qtde_cpu,
            'qtde_ram_mb': self.qtde_ram_mb,
            'redes': [rede.to_dict() for rede in self.redes],
            'id_vmm': self.id_vmm,
            'status': self.status.value,
            'no_regiao': self.no_regiao,
            'ansible': [self.dados_ansible[dados_ansible].to_dict()
                        for dados_ansible in self.dados_ansible]
        }

        if self.to_json_dados_completos:
            dict_objeto['discos_adicionais'] = [self.discos_adicionais[disco_adicional].to_dict()
                                                for disco_adicional in self.discos_adicionais]

        return dict_objeto



"""
Módulo com funções úteis de operação
"""

import sys
import os
import platform
import json
from vmm_manager.util.msgs import imprimir_ok, imprimir_erro
from vmm_manager.util.msgs import formatar_msg_erro, finalizar_com_erro
from vmm_manager.util.msgs import get_str_data_formatada, imprimir_acao_corrente
from vmm_manager.infra.comando import Comando


def __confirmar_acao_usuario(servidor_acesso=None, agrupamento=None, nuvem=None,
                             ocultar_progresso=False):
    resposta = None
    while resposta not in ['s', 'n']:
        resposta = input('Deseja executar? (s/n): ')
        if resposta == 'n':
            if servidor_acesso and agrupamento and nuvem:
                liberar_lock(servidor_acesso, agrupamento,
                             nuvem, ocultar_progresso)
            print('Ação cancelada pelo usuário.')
            sys.exit(0)


def confirmar_acao_usuario_sem_lock():
    __confirmar_acao_usuario()


def confirmar_acao_usuario_com_lock(servidor_acesso, agrupamento, nuvem):
    __confirmar_acao_usuario(servidor_acesso, agrupamento, nuvem)


def validar_retorno_operacao_sem_lock(status, msg, ocultar_progresso):
    if status:
        imprimir_ok(ocultar_progresso)
    else:
        imprimir_erro(ocultar_progresso)
        finalizar_com_erro(msg)


def validar_retorno_operacao_com_lock(status, msg, servidor_acesso,
                                      agrupamento, nuvem, ocultar_progresso):
    if status:
        imprimir_ok(ocultar_progresso)
    else:
        imprimir_erro(ocultar_progresso)
        liberar_lock(servidor_acesso, agrupamento, nuvem, ocultar_progresso)
        finalizar_com_erro(msg)


def adquirir_lock(servidor_acesso, agrupamento, nuvem, ocultar_progresso):
    imprimir_acao_corrente('Adquirindo lock', ocultar_progresso)

    cmd = Comando('adquirir_lock',
                  lockfile=servidor_acesso.get_caminho_lockfile(
                      agrupamento,
                      nuvem),
                  pid=os.getpid(),
                  host=platform.node(),
                  agrupamento=agrupamento,
                  nuvem=nuvem,
                  data=get_str_data_formatada('%d/%m/%Y às %H:%M:%S.%f'))

    status, retorno_cmd = cmd.executar(servidor_acesso)
    if status:
        dados_lock = json.loads(retorno_cmd)
        if not dados_lock.get('Sucesso'):
            status = False
            retorno_cmd = f"O processo {dados_lock.get('PIDProcesso')}, " \
                f"iniciado em { dados_lock.get('DataLock')} " \
                f"no servidor {dados_lock.get('HostLock')}, " \
                f"já está manipulando o agrupamento {dados_lock.get('Agrupamento')} " \
                f"na nuvem {dados_lock.get('Nuvem')}."

    validar_retorno_operacao_sem_lock(status, retorno_cmd, ocultar_progresso)


def liberar_lock(servidor_acesso, agrupamento, nuvem, ocultar_progresso):
    imprimir_acao_corrente('Liberando lock', ocultar_progresso)

    cmd = Comando('liberar_lock',
                  lockfile=servidor_acesso.get_caminho_lockfile(
                      agrupamento,
                      nuvem))

    status, _ = cmd.executar(servidor_acesso)
    if status:
        imprimir_ok(ocultar_progresso)
    else:
        imprimir_erro(ocultar_progresso)
        print(formatar_msg_erro(
            'Erro ao liberar o lock: você terá que excluir o arquivo manualmente.'))



"""
Módulo que realiza o parser de um inventário local
"""
import os
import yamale
from vmm_manager.entidade.inventario import Inventario
from vmm_manager.entidade.vm import VM
from vmm_manager.entidade.vm_rede import VMRede
from vmm_manager.scvmm.enums import SCDiskBusType, SCDiskSizeType
from vmm_manager.entidade.vm_disco import VMDisco
from vmm_manager.scvmm.scregion import SCRegion


class ParserLocal:
    __ARQUIVO_SCHEMA = '../includes/schema.yaml'
    __YAML_PARSER = 'ruamel'

    @staticmethod
    def __get_schema_yaml():
        return yamale.make_schema(
            os.path.join(
                os.path.dirname(__file__), ParserLocal.__ARQUIVO_SCHEMA),
            parser=ParserLocal.__YAML_PARSER)

    @staticmethod
    def __get_discos_adicionais(nome_vm, dict_discos_adicionais):
        discos_adicionais = []
        for item in dict_discos_adicionais or {}:
            arquivo = item.get('arquivo')

            if arquivo in [disco_adicional.arquivo for disco_adicional in discos_adicionais]:
                raise ValueError(
                    f"Disco '{arquivo}' referenciado mais de uma vez "
                    f"para a VM '{nome_vm}'.")

            disco_adicional = VMDisco(
                SCDiskBusType(item.get('tipo')),
                item.get('arquivo'),
                item.get('tamanho_mb'),
                SCDiskSizeType(item.get('tamanho_tipo')),
                item.get('caminho'))

            discos_adicionais.append(disco_adicional)

        return discos_adicionais

    def __init__(self, arquivo_inventario):
        self.__arquivo_inventario = arquivo_inventario
        self.__inventario = None

    def __validar_arquivo_yaml(self):
        if not os.path.isfile(self.__arquivo_inventario):
            raise ValueError('Arquivo de inventário não encontrado.')
        if os.stat(self.__arquivo_inventario).st_size == 0:
            raise ValueError('Arquivo de inventário vazio.')

    def __montar_inventario(self, dados_inventario,
                            filtro_nome_vm=None, filtro_dados_completos=True):
        nomes_vm = []
        self.__inventario = Inventario(
            dados_inventario['agrupamento'], dados_inventario['nuvem'])

        for maquina_virtual in dados_inventario['vms']:
            nome_vm = maquina_virtual.get('nome').upper()
            if nome_vm in nomes_vm:
                raise ValueError(
                    f'VM {nome_vm} referenciada mais de uma vez no inventário.')
            nomes_vm.append(nome_vm)

            # filtrando vms: melhoria no desempenho
            if filtro_nome_vm and nome_vm != filtro_nome_vm:
                continue

            vm_redes = []
            nomes_redes = []
            for rede_vm in maquina_virtual.get('redes', dados_inventario.get('redes_padrao', [])):
                nome_rede = rede_vm.get('nome')

                if nome_rede in nomes_redes:
                    raise ValueError(
                        f"Rede '{nome_rede}' referenciada mais de uma vez para a VM '{nome_vm}'.")

                nomes_redes.append(nome_rede)
                vm_redes.append(
                    VMRede(nome_rede, rede_vm.get('principal', False)))

            self.__inventario.vms[nome_vm] = VM(
                nome_vm,
                maquina_virtual.get('descricao'),
                maquina_virtual.get(
                    'imagem', dados_inventario.get('imagem_padrao', None)),
                maquina_virtual.get('regiao', SCRegion.REGIAO_PADRAO),
                maquina_virtual.get(
                    'qtde_cpu', dados_inventario.get('qtde_cpu_padrao', None)),
                maquina_virtual.get(
                    'qtde_ram_mb', dados_inventario.get('qtde_ram_mb_padrao', None)),
                vm_redes
            )
            self.__inventario.vms[nome_vm].extrair_dados_ansible_dict(
                maquina_virtual.get('ansible'))

            # Obtendo dados adicionais
            if filtro_dados_completos:
                # discos
                self.__inventario.vms[nome_vm].add_discos_adicionais(
                    ParserLocal.__get_discos_adicionais(nome_vm,
                                                        maquina_virtual.get('discos_adicionais')))

    def __carregar_yaml(self):
        return yamale.make_data(self.__arquivo_inventario,
                                parser=ParserLocal.__YAML_PARSER)

    def get_inventario(self, servidor_acesso, filtro_nome_vm=None, filtro_dados_completos=True):
        if not self.__inventario:
            try:
                self.__validar_arquivo_yaml()
                dados_yaml = self.__carregar_yaml()
                parser = ParserLocal.__get_schema_yaml()

                yamale.validate(parser, dados_yaml, strict=True)

                self.__montar_inventario(
                    dados_yaml[0][0], filtro_nome_vm, filtro_dados_completos)
                self.__inventario.validar(servidor_acesso)
            except (SyntaxError, ValueError) as ex:
                return False, str(ex)

        return True, self.__inventario
